from ._version import (__version__, __updated__)

__copyright__     = "Copyright (C) 2013-2015 European Commission (JRC)"
__license__       = "EUPL 1.1+"

__title__ = "pandalone"
__summary__ = "Collection of utilities for working with hierarchical data with relocatable paths."
__uri__ = "https://github.com/pandalone/pandalone"
