#!/usr/bin/env python
#
# Copyright 2014-2015 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""Authoritative project's version-info"""
__version__ = version = "0.2.2"  # PEP 440 versioning. Also update README.rst's.
__updated__ = "2017-02-07 04:05:41"
